import os
try:
    import yaml
except ModuleNotFoundError:
    yaml = None

from LordUtils.json_encoder import JEncoder
from LordPath import save_file, read_file
from LordPath.opener import Opener


def load_yaml(y, default=None):
    if yaml is None:
        raise ModuleNotFoundError(' PyYAML - pip install PyYAML ')

    if os.path.isfile(y):
        y = read_file(y, None)
    if y is None:
        return default

    _yaml = yaml.safe_load(y)
    if isinstance(_yaml, dict):
        return _yaml
    elif isinstance(_yaml, list):
        return _yaml
    else:
        return default


def save_yaml(obj, path):
    if yaml is None:
        raise ModuleNotFoundError(' PyYAML - pip install PyYAML ')

    _y = yaml.dump(obj)
    return save_file(path, _y)


class YamlOpener(Opener):
    compatible_endings = ['yaml', 'yml']

    @staticmethod
    def load(file, default=None) -> dict:
        return load_yaml(file, default=None)

    @staticmethod
    def save(obj, file):
        return save_yaml(obj, file)






# compact
file_type_to_endings = {
    'image': ['png', 'jpeg', 'jpg', 'gif', 'svg'],
    'video': ['mkv', 'mp4', 'mpeg'],
    'audio': ['mp3', 'aac', 'flac', 'm3u', 'm4b', 'wav'],
    'usenet': ['nzb'],
    'text': ['txt'],
    'scene': ['nfo'],
    'archive': ['rar', 'zip', '7z', 'tar'],
    'documents': ['pdf', 'docs'],
    'code': ['py', 'js', 'html', 'css'],
    'ebook': ['epub', 'mobi'],
    'meta': ['json', 'yaml', 'yml', 'xml']
}

# flat
ending_to_file_type = {}
for type, file_type_list in file_type_to_endings.items():
    for file_type in file_type_list:
        ending_to_file_type[file_type] = type



import pickle
from pathlib import Path
from .. import _path_convert

from LordPath import path_check, mkdir
from .base import Opener

# pyobj
# https://www.techcoil.com/blog/how-to-save-and-load-objects-to-and-from-file-in-python-via-facilities-from-the-pickle-module/
# https://www.journaldev.com/15638/python-pickle-example


def save_obj(obj, file: str | Path):
    file: Path = path_check(file)
    mkdir(file.parent)

    with file.open('wb') as f:
        pickle.dump(obj, f)


def load_obj(file: Path | str):
    file = _path_convert(file)
    with file.open('rb') as f:
        data = pickle.load(f)
        return data


class PickleOpener(Opener):
    compatible_endings = ['pickle']

    @staticmethod
    def load(file, default=None) -> dict:
        return load_obj(file)

    @staticmethod
    def save(obj, file):
        return save_obj(obj, file)



try:
    import toml
except ModuleNotFoundError:
    toml = None
from .base import Opener


class TomlOpener(Opener):
    compatible_endings = ['toml']

    @staticmethod
    def load(file, default=None) -> dict:
        if toml is None:
            raise ModuleNotFoundError(' TOML - pip install toml ')
        with open(file, 'r') as f:
            return toml.load(f)

    @staticmethod
    def save(obj, file):
        with open(file, 'r') as f:
            return toml.dump(obj, file)


from os import PathLike
from pathlib import Path


from .base import Opener
from .. import read_file, save_file
from . import yaml, json, pickle, toml


def get_opener(file_ending):
    for cls in Opener.__subclasses__():
        if cls.check(file_ending):
            return cls
    return None


def open_by_type(path: Path, ending=None, default=None):
    if ending is None:
        ending = path.suffix
    opener_cls = get_opener(ending)
    if opener_cls is None:
        return default
    return opener_cls.load(path)


def save_by_type(obj, path, ending=None, default=None):
    if ending is None:
        ending = path.suffix
    opener_cls = get_opener(ending)
    if opener_cls is None:
        return default
    return opener_cls.save(obj, path)


class TextOpener(Opener):
    compatible_endings = ['txt']

    @staticmethod
    def load(file, default=None):
        return read_file(file, default=None)

    @staticmethod
    def save(text, file):
        return save_file(file, text)






from pathlib import Path


class PathDescriptor:
    def __set_name__(self, owner, name):
        self.public_name = name
        self.private_name = '_' + name

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        value = getattr(instance, self.private_name, None)
        return value

    def __set__(self, instance, value):
        if isinstance(value, str):
            value = Path(value)
        elif isinstance(value, (int, float)):
            value = Path(str(value))
        setattr(instance, self.private_name, value)



import shutil
from pathlib import Path
# pip install regex
import regex as re
from typing import Union

# File Path
patternFileName = re.compile(r'[\/\\\?\<\>\|\:\'\"\*,]')  # / \ ? < > | : ' " * ,

# configs
encodes = ["utf-8-sig", "utf-8", "utf-16", "cp1252", "iso8859-1"]
list_dir_filter = ['.DS_Store', '.git']


def _path_convert(path: str | Path) -> Path:
    if isinstance(path, str):
        return Path(path)
    else:
        return path


def listdir(path, *, only_files=False, only_folders=False, as_fullpath=False):
    path = _path_convert(path)

    file_list = [f for f in path.iterdir() if f.name not in list_dir_filter]

    if only_files:
        file_list = [f for f in file_list if f.is_file()]
    if only_folders:
        file_list = [f for f in file_list if f.is_dir()]

    return file_list


def mkdir(*path):
    for p in path:
        p = _path_convert(p)
        p.mkdir(parents=True, exist_ok=True)


def makedirs(*path):
    mkdir(*path)


def clear_empty_folder(path, recursive=False):
    path_obj = Path(path)

    # Find all subdirectories of the given path that don't match any filter and iterate over them.
    folders = [subdir for subdir in path_obj.iterdir() if
               subdir.is_dir() and not any(str(subdir).endswith(i) for i in list_dir_filter)]

    # Check if each subdirectory is empty and delete it, if empty.
    for folder in folders:
        if recursive:
            clear_empty_folder(folder, True)
        files = list(folder.iterdir())
        files = [f for f in files if f.name not in ['.DS_Store']]
        if not any(files):
            shutil.rmtree(folder)


def file_name_check(file_name):
    return patternFileName.sub('', file_name)


def path_check(path: Path | str) -> Path:
    """
    Replace illegal characters in file / folder path
    """
    if not isinstance(path, str):
        path = str(path)

    folders = re.split(r"[\\/]", path)
    for i in range(0, len(folders)):
        f = folders[i]
        if i == 0:
            if len(f) == 2 and ":" in f:
                continue
        f = str(file_name_check(f))
        folders[i] = f

    return Path(path)


# save load Files

typical_readable_file_endings = {
    '.txt',
    '.nfo',
    '.html',
    '.md'
}


def read_file(file: Union[str, list, Path], default=None):
    """
    :param file: file to be loaded. Specification as list to indicate alternative files
    :param default: default return
    """
    if isinstance(file, list):
        for f in file:
            _text = read_file(f)
            if _text:
                return _text

    file = Path(file)
    if not file.exists():
        return default

    for enc in encodes:
        try:
            with file.open("r", encoding=enc) as f:
                return f.read()
        except:
            pass

    return default


def typical_readable_file_type(filename):
    path = Path(filename)
    return path.suffix in typical_readable_file_endings


# file

def save_file(file, text):

    file: Path = path_check(file)
    dir_path = file.parent
    if not dir_path.exists():
        makedirs(dir_path)
    with file.open("w", encoding=encodes[0]) as f:
        f.write(text)
    return True



from pathlib import Path

from LordPath import read_file, typical_readable_file_type
from LordPath import listdir, makedirs
from LordPath.opener import open_by_type
from LordPath.dataset import ending_to_file_type
from LordUtils.decorators import CachedProperty

# LordThings
from LordUtils import fn_match_value


class FileClass(object):
    type = 'file'

    def __init__(self, path: str | Path):
        self.path = Path(path)
        self.file_type = ending_to_file_type.get(self.ending, 'unknown')

        self._open = None
        self._text = None

    def open(self):
        if self._open is None:
            self._open = open_by_type(self.path, ending=self.ending)
        return self._open

    @property
    def filename(self):
        return self.path.name

    @property
    def name(self):
        return self.path.stem

    @property
    def extension(self):
        """Alias zu ending"""
        return self.ending

    @property
    def ending(self):
        """Alias zu ending"""
        return self.path.suffix.replace('.', '')

    @property
    def full_name(self):
        return self.path.name

    @property
    def data(self):
        return self.open()

    @property
    def text(self):
        if self._text is None:
            self._text = self.get_text()
        return self._text

    def get_text(self, only_typical=False):
        if only_typical is False or typical_readable_file_type(self.filename):
            r = read_file(self.path)
        else:
            r = '...'
        return r

    def to_json(self):
        return {
            'path': '--CENSORED--',
            'filename': self.filename,
            'name': self.name,
            'text': self.get_text(only_typical=True),
            'ending': self.ending,
            'file_type': self.file_type,
            'type': self.type,
            'data': self.data,
            '__class__': self.__class__.__name__
        }

    def move(self, target_path):
        target_path = Path(target_path)
        if not target_path.exists():
            makedirs(target_path)

        new_path = target_path / self.filename
        self.path = self.path.rename(new_path)

    def __str__(self):
        return self.filename

    def __getitem__(self, key):  # => self[key]
        return self.__getattribute__(key)


class FolderClass(object):
    type = "folder"

    def __init__(self, path):
        self.path = Path(path)

    @property
    def name(self):
        return self.path.name

    @CachedProperty
    def elements(self):
        elements = []
        _p = self.path
        for f in listdir(_p):
            f = create_from_path(f)
            elements.append(f)
        return elements

    @property
    def full_name(self):
        return self.name

    @property
    def all_files(self):
        for f in self.elements:
            if isinstance(f, FolderClass):
                yield from f.all_files
            else:
                yield f

    @property
    def files(self):
        return [f for f in self.elements if isinstance(f, FileClass)]

    @property
    def folders(self):
        return [f for f in self.elements if isinstance(f, FolderClass)]

    def to_json(self):
        return {
            # todo make censor setting for this
            'path': '--CENSORED--',
            'name': self.name,
            'type': self.type,
            '__class__': self.__class__.__name__
        }

    def search(self, *q: str | list):
        # check files
        for f in self.files:
            if fn_match_value(str(f), q):
                yield f
        for f in self.folders:
            yield from f.search(*q)

    def search_one(self, *q):
        for f in self.search(*q):
            return f

    def move(self, target_path):
        target_path = Path(target_path)
        if not target_path.exists():
            makedirs(target_path)

        new_path = target_path / self.name
        self.path = self.path.rename(new_path)

    def __str__(self):
        return self.name

    def __iter__(self):
        return iter(self.elements)

    def __reversed__(self):
        return reversed(self.elements)

    def __getitem__(self, key):  # => self[key]
        return self.__getattribute__(key)


class VirtualFolderClass(object):
    type = "virtual-folder"

    def __init__(self, name=None):
        self.name = name
        self.elements: list[FileClass, FolderClass] = []

    def add(self, *elements: FolderClass | FileClass):
        for e in elements:
            self.elements.append(e)

    def remove(self, *elements: FolderClass | FileClass):
        for e in elements:
            if e in self.elements:
                self.elements.remove(e)

    @classmethod
    def init_from_folder(cls, path):
        path = Path()
        name = path.name
        v = cls(name)
        for f in path.iterdir():
            f = create_from_path(f)
            v.add(f)

    @property
    def files(self):
        return [f for f in self.elements if isinstance(f, FileClass)]

    @property
    def folders(self):
        return [f for f in self.elements if isinstance(f, FolderClass)]

    @property
    def all_files(self):
        for f in self.elements:
            if isinstance(f, FolderClass):
                yield from f.all_files
            else:
                yield f

    def to_json(self):
        return {
            'name': self.name,
            'type': self.type,
            '__class__': self.__class__.__name__,
            'files': self.files,
            'folders': self.folders
        }

    def search(self, *q: str):
        # check files
        for f in self.files:
            if fn_match_value(str(f), q):
                yield f
        for f in self.folders:
            yield from f.search(*q)

    def search_one(self, *q):
        for f in self.search(*q):
            return f

    def move(self, target_path):
        for e in self.elements:
            e.move(target_path)

    def __iter__(self):
        return iter(self.elements)

    def __reversed__(self):
        return reversed(self.elements)


vffc_type = FileClass | FolderClass | VirtualFolderClass
# vffc => virtual file folder class


def create_from_path(path: str | Path):
    path = Path(path)

    if path.is_file():
        return FileClass(path)
    elif path.is_dir():
        return FolderClass(path)
    else:
        return None



import os
import json

from LordUtils.json_encoder import JEncoder

from LordPath import save_file, read_file
from .base import Opener


def load_json(j, default=None):
    if os.path.isfile(j):
        j = read_file(j, None)
    _strip = j.strip()
    if not (_strip.endswith(('}', ']')) and _strip.startswith(('{', '['))):
        return default
    if j is None:
        return default

    _json = json.loads(j)
    return _json


def save_json(obj, path, cls=None):
    if cls is None:
        cls = JEncoder
    _j = json.dumps(obj, indent=2, cls=cls)
    return save_file(path, _j)


class JsonOpener(Opener):
    compatible_endings = ['json']

    @staticmethod
    def load(file, default=None) -> dict:
        return load_json(file, default=None)

    @staticmethod
    def save(obj, file):
        return save_json(obj, file)


__all__ = [
    'load_json'
]



from abc import ABC, abstractmethod


class Opener(ABC):
    compatible_endings = []

    @classmethod
    def check(cls, ending):
        return ending in cls.compatible_endings

    @staticmethod
    @abstractmethod
    def load(file):
        pass

    @staticmethod
    @abstractmethod
    def save(obj, file):
        pass